import argparse

from . import compile

def main() -> None:
    parser = argparse.ArgumentParser(description="Run API query.")
    parser.add_argument("file", type=str,
                        help="Query file.")
    parser.add_argument("--compile-only", action='store_true',
                        help="Only compile the query and print resulting Python code to stdout.")
    parser.add_argument("--max-concurrent", type=int, default=1,
                        help="Max number of concurrent statement executions allowed.")
    parser.add_argument("--http-rate-limit", type=int, default=1,
                        help="Max number of HTTP queries per second.")
    parser.add_argument("--http-retry-count", type=int, default=1,
                        help="Max number of times to retry HTTP queries.")
    parser.add_argument("--http-base-delay", type=float, default=1.0,
                        help="Base delay between HTTP query retries.")
    parser.add_argument("--http-max-delay", type=float, default=10.0,
                        help="Max delay between HTTP query retries.")
    parser.add_argument("--log-level", type=str, default='info',
                        choices=['debug', 'info', 'warning', 'error', 'fatal'],
                        help="Log level.")
    args = parser.parse_args()

    with open(args.file, 'rt') as f:
        source = compile(f.read())

    if args.compile_only:
        print(source)
    else:
        exec(source, {'__name__': '__main__'})

if __name__ == '__main__':
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Sequence, NamedTuple

class Statement(ABC):
    __slots__: Sequence[str] = tuple()

    @abstractmethod
    def compile(self, indent: str = '') -> str:
        pass

class Identifier(str):
    __slots__: Sequence[str] = tuple()

    def __repr__(self) -> str:
        return str(self)

class OutputIdentifier(Identifier):
    __slots__: Sequence[str] = tuple()

    def __repr__(self) -> str:
        return f'OutputIdentifier({super(Identifier, self).__repr__()})'

class OptionalOutputIdentifier(Identifier):
    __slots__: Sequence[str] = tuple()

    def __repr__(self) -> str:
        return f'OptionalOutputIdentifier({super(Identifier, self).__repr__()})'

class _Placeholder:
    __slots__: Sequence[str] = tuple()

    def __str__(self) -> str:
        return 'Placeholder'

    def __repr__(self) -> str:
        return 'None'
Placeholder = _Placeholder()

TemplateValue = int | float | bool | str | None | Identifier | _Placeholder

RequestTemplateObject = dict[str, 'RequestTemplate']
RequestTemplateArray = list['RequestTemplate']
RequestTemplate = TemplateValue | RequestTemplateObject | RequestTemplateArray

ResponseTemplateObject = dict[str, 'ResponseTemplate']
ResponseTemplateArray = list['ResponseTemplate']

class ResponseTemplateArrayHandler(NamedTuple):
    template: 'ResponseTemplate'
    handler: list[Statement]

    @classmethod
    def make(cls, args: tuple['ResponseTemplate', list[Statement]]) -> ResponseTemplateArrayHandler:
        return cls(*args)

ResponseTemplate = TemplateValue | OutputIdentifier | ResponseTemplateObject | ResponseTemplateArray | \
                   ResponseTemplateArrayHandler

@dataclass(slots=True)
class PythonStatement(Statement):
    source: str

    @classmethod
    def from_assignment(cls, key: str, value: str) -> PythonStatement:
        if value:
            return cls(source=f"{key} = f'''{value}'''")
        else:
            return cls(source=f"{key} = os.environ.get('{key}', '')")

    def compile(self, indent: str = '') -> str:
        return (
            f'{indent}async with __stmt_limit:\n'
            f'{indent}    __lib.LOG.info("Running python: " {repr(self.source.strip())})\n'
            f'{indent}    {self.source.strip()}\n'
        )

@dataclass(slots=True)
class ShellStatement(Statement):
    source: str

    def compile(self, indent: str = '') -> str:
        return (
            f'{indent}async with __stmt_limit:\n'
            f'{indent}    __lib.LOG.info("Running shell: " {repr(self.source.strip())})\n'
            f'{indent}    await __lib.run_shell(f{repr(self.source.strip())})\n'
        )

@dataclass(slots=True)
class HTTPStatement(Statement):
    method: str
    url: str
    headers: dict[str, str]
    body: RequestTemplateObject | RequestTemplateArray | str | None
    response: ResponseTemplateObject | ResponseTemplateArray | OutputIdentifier

    def _output_headers(self, indent: str = '') -> str:
        if not self.headers: return '{}'
        return (
             '{\n' +
             ''.join(f'{indent}    {repr(key)}: f{repr(value)}\n' for key, value in self.headers.items()) +
            f'{indent}}}'
        )

    def _output_response_object_element(
            self, sfx: str, path: str, key: str, value: ResponseTemplate, indent: str) -> str:
        new_sfx = str(id(value))
        if isinstance(value, OptionalOutputIdentifier):
            return (
                f'{indent}__p{new_sfx} = __p{sfx}.get({repr(key)}, None)\n' +
                self._output_response_general(value, new_sfx, f'{path}.{key}', indent) +
                f'{indent}del __p{new_sfx}\n'
            )
        return (
            f'{indent}assert {repr(key)} in __p{sfx}, f"{path} in response has no key " + repr({repr(key)})\n'
            f'{indent}__p{new_sfx} = __p{sfx}[{repr(key)}]\n' +
            self._output_response_general(value, new_sfx, f'{path}.{key}', indent) +
            f'{indent}del __p{new_sfx}\n'
        )

    def _output_response_object(
            self, response: ResponseTemplateObject, sfx: str, path: str, indent: str) -> str:
        return (
            f'{indent}assert isinstance(__p{sfx}, dict), f"{path} in response is not an object"\n' +
            ''.join(self._output_response_object_element(sfx, path, k, r, indent)
                    for k, r in response.items())
        )

    def _output_response_array_element(
            self, sfx: str, path: str, key: int, value: ResponseTemplate, indent: str) -> str:
        r = self._output_response_general(value, str(id(value)), f'{path}[{key}]', indent)
        if not r:
            return ''
        return (
            f'{indent}assert len(__p{sfx}) > {key}, f"{path} in response has fewer than {key + 1} elements"\n'
            f'{indent}__p{id(value)} = __p{sfx}[{repr(key)}]\n'
            f'{r}'
            f'{indent}del __p{id(value)}\n'
        )

    def _output_response_array(
            self, response: ResponseTemplateArray, sfx: str, path: str, indent: str) -> str:
        return (
            f'{indent}assert isinstance(__p{sfx}, list), f"{path} in response is not an array"\n' +
            ''.join(self._output_response_array_element(sfx, path, i, r, indent) for i, r in enumerate(response))
        )

    def _output_response_array_handler(
            self, response: ResponseTemplateArrayHandler, sfx: str, path: str, indent: str) -> str:
        return (
            f'{indent}assert isinstance(__p{sfx}, list), f"{path} in response is not an array"\n\n'
            f'{indent}async def __f{sfx}(__p{sfx}_e):\n' +
            f'{indent}    __lib.resolve_imports()\n' +
            self._output_response_general(response.template, f'{sfx}_e', f'{path}[]', indent + '    ') +
            ''.join(f'{stmt.compile(indent=indent + "    ")}\n' for stmt in response.handler) +
            f'\n'
            f'{indent}await asyncio.gather(*{{__f{sfx}(__p{sfx}_e) for __p{sfx}_e in __p{sfx}}})\n'
            f'{indent}del __f{sfx}\n'
        )


    def _output_response_general(
            self, response: ResponseTemplate, sfx: str, path: str, indent: str = '') -> str:
        if isinstance(response, _Placeholder):
            return ''
        if isinstance(response, OutputIdentifier | OptionalOutputIdentifier):
            return f'{indent}{response} = __p{sfx}\n'
        elif isinstance(response, str | int | float | bool | None | Identifier):
            return (
                f'{indent}assert __p{sfx} == {repr(response)}, "{path} in response not equal to " + '
                        f'repr({repr(response)})\n'
            )
        elif isinstance(response, dict):
            return self._output_response_object(response, sfx, path, indent)
        elif isinstance(response, list):
            return self._output_response_array(response, sfx, path, indent)
        else:
            return self._output_response_array_handler(response, sfx, path, indent)

    def _output_response(self, sfx: str, indent: str = '') -> str:
        return self._output_response_general(self.response, sfx, 'RESPONSE', indent)

    def compile(self, indent: str = '') -> str:
        sfx = f'{id(self)}'
        body_value = repr(self.body) if not isinstance(self.body, str) else f'f"{self.body}"'
        data_arg = (
            f'data=__body{sfx}'
            if isinstance(self.body, str) else
            f'json=__body{sfx}'
            if self.body is not None else
            ''
        )
        response_type = 'text' if isinstance(self.response, OutputIdentifier) else 'json'
        return (
                f'{indent}__headers{sfx} = {self._output_headers(indent)}\n'
                f'{indent}__body{sfx} = {body_value}\n'
                f'{indent}__url{sfx} = f{repr(self.url)}\n'
                f'{indent}__retry_delay{sfx} = __http_retry.base_delay\n'
                f'{indent}for __retry{sfx} in range(1, __http_retry.retry_count + 1):\n'
                f'{indent}    await __http_limit.get_token()\n'
                f'{indent}    try:\n'
                f'{indent}        async with __stmt_limit:\n'
                f'{indent}                __lib.LOG.info(f"Running {self.method} request to: {{__url{sfx}}}," '
                                                       f'f"attempt {{__retry{sfx}}} / {{__http_retry.retry_count}}")\n'
                f'{indent}                __lib.LOG.debug(f"Headers: {{__headers{sfx}}}, body: {{__body{sfx}}}")\n'
                f'{indent}                async with __session.{self.method.lower()}(__url{sfx}, '
                                                  f'headers=__headers{sfx}, {data_arg}) as __response{sfx}:\n'
                f'{indent}                    __p{sfx} = await __response{sfx}.{response_type}()\n'
                f'{indent}        __lib.LOG.debug(f"Response for {{__url{sfx}}}: {{__p{sfx}}}")\n'
                f'{self._output_response(sfx, indent + " " * 8)}'
                f'{indent}        del __p{sfx}, __response{sfx}\n'
                f'{indent}        break\n'
                f'{indent}    except Exception as e:\n'
                f'{indent}        __lib.LOG.warning(f"{self.method} request to {{__url{sfx}}} failed "'
                                                  f'f"({{__retry{sfx}}} / {{__http_retry.retry_count}}): {{e}}")\n'
                f'{indent}    if __retry{sfx} != __http_retry.retry_count:\n'
                f'{indent}        await asyncio.sleep(__retry_delay{sfx})\n'
                f'{indent}        __retry_delay{sfx} = min(2 * __retry_delay{sfx}, __http_retry.max_delay)\n'
                f'{indent}else:\n'
                f'{indent}    raise RuntimeError("{self.method} request to {{__url{sfx}}} failed.")\n'
                f'{indent}del __headers{sfx}, __body{sfx}, __url{sfx}, __retry{sfx}, __retry_delay{sfx}\n'
        )



from typing import Iterator

from .statement import Statement

PREAMBLE = '''
import asyncio
import aiohttp
import api_query.lib as __lib

async def __run_query(__session: aiohttp.ClientSession,
                      __stmt_limit: __lib.ConcurrencyLimit,
                      __http_limit: __lib.RateLimit,
                      __http_retry: __lib.RetryConfig) -> None:
    __lib.resolve_imports()
'''

POSTAMBLE = '''

async def __query_async(max_concurrent: int, http_rate_limit: int, http_retry: __lib.RetryConfig) -> None:
    async with aiohttp.ClientSession() as session:
        try:
            await __run_query(session,
                              __lib.ConcurrencyLimit(max_concurrent),
                              __lib.RateLimit(http_rate_limit),
                              http_retry)
        except Exception as e:
             __lib.LOG.error(str(e))

def __query(max_concurrent: int, http_rate_limit: int, http_retry: __lib.RetryConfig) -> None:
    asyncio.run(__query_async(max_concurrent, http_rate_limit, http_retry))

if __name__ == "__main__":
    import argparse
    __parser = argparse.ArgumentParser("Run API query.")
    __parser.add_argument("--max-concurrent", type=int, default=1,
                          help="Max number of concurrent statement executions allowed.")
    __parser.add_argument("--log-level", type=str, default='info',
                          choices=['debug', 'info', 'warning', 'error', 'fatal'],
                          help="Log level.")
    __parser.add_argument("--http-rate-limit", type=int, default=1,
                          help="Max number of HTTP queries per second.")
    __parser.add_argument("--http-retry-count", type=int, default=1,
                          help="Max number of times to retry HTTP queries.")
    __parser.add_argument("--http-base-delay", type=float, default=1.0,
                          help="Base delay between HTTP query retries.")
    __parser.add_argument("--http-max-delay", type=float, default=10.0,
                          help="Max delay between HTTP query retries.")
    __args, __ = __parser.parse_known_args()

    __lib.set_log_level(__args.log_level)
    __query(__args.max_concurrent, __args.http_rate_limit,
            __lib.RetryConfig(__args.http_retry_count, __args.http_base_delay, __args.http_max_delay))
'''

def compile(statements: Iterator[Statement]) -> Iterator[str]:
    yield PREAMBLE
    for statement in statements:
        yield statement.compile(indent='    ')
    yield POSTAMBLE

if __name__ == '__main__':
    import sys
    from .lexer import lex
    from .parser import parse
    with open(sys.argv[1], 'rt') as f:
        for x in compile(parse(lex(f.read()))):
            print(x)



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from io import StringIO
import re
import shlex
import tokenize
from typing import Any, Callable, NamedTuple

from funcparserlib.lexer import Token, LexerError

class Position(NamedTuple):
    line: int
    char: int
    offset: int

@dataclass(frozen=True, slots=True)
class Slice:
    raw: str
    start: Position
    # end is inclusive
    end: Position

    def __add__(self, s: Slice) -> Slice:
        assert self.raw == s.raw, 'Cannot add slices of different strings'
        assert self.end.offset == s.start.offset - 1, 'Cannot add non-contiguous slices'
        return Slice(self.raw, self.start, s.end)

    def __getitem__(self, val: int | slice) -> Slice:
        length = self.end.offset - self.start.offset + 1
        if isinstance(val, int):
            start = end = val if val >= 0 else length + val
        else:
            start = 0 if val.start is None else val.start
            end = length if val.stop is None else val.stop
            if start < 0:
                start += length
            elif start >= length:
                start = length - 1
            if end < 0:
                end += length
            elif end >= length:
                end = length
            end -= 1

        assert 0 <= start <= end <= length, "only non-empty slices are supported"

        if self.start.line == self.end.line:
            return Slice(
                    self.raw,
                    Position(self.start.line, self.start.char + start, self.start.offset + start),
                    Position(self.start.line, self.start.char + end, self.start.offset + end))
        else:
            offset_start = self.start.offset + start
            offset_end = self.start.offset + end

            nl_to_start = self.raw.count('\n', self.start.offset, offset_start)
            if nl_to_start == 0:
                start_pos = self.start.char + start
            else:
                start_pos = offset_start - self.raw.rfind('\n', self.start.offset, offset_start)
            if self.start.line + nl_to_start == self.end.line:
                nl_to_end = nl_to_start
            else:
                nl_to_end = nl_to_start + self.raw.count('\n', offset_start, offset_end)
            if nl_to_end == 0:
                end_pos = self.start.char + end
            else:
                end_pos = offset_end - self.raw.rfind('\n', self.start.offset, offset_end)
            return Slice(
                    self.raw,
                    Position(self.start.line + nl_to_start, start_pos, self.start.offset + start),
                    Position(self.start.line + nl_to_end, end_pos, self.start.offset + end))

    def get(self) -> str:
        return self.raw[self.start.offset : self.end.offset + 1]

    def to_eol(self) -> tuple[Slice | None, Slice | None]:
        index = self.raw.find('\n', self.start.offset, self.end.offset)
        if index == self.start.offset:
            return None, self
        elif self.start.offset < index < self.end.offset:
            pivot = index - self.start.offset
            return self[:pivot], self[pivot:]
        else:
            return self, None

    def consume_eol(self) -> tuple[Slice | None, Slice | None]:
        assert self.raw[self.start.offset] == '\n'
        if self.start == self.end:
            return self, None
        else:
            return self[:1], self[1:]

    @classmethod
    def from_str(cls, raw: str) -> Slice | None:
        length = len(raw)
        if length == 0:
            return None
        lines = raw.count('\n', 0, length - 1) + 1
        if lines == 1:
            end = Position(1, length, length - 1)
        else:
            end = Position(lines, (length - 1) - raw.rindex('\n', 0, length - 1), length - 1)
        return cls(raw, Position(1, 1, 0), end)

TokenMatcher = Callable[[Slice, Slice | None], tuple[Slice, Slice | None]]

def match_to_exception(
    fn: Callable[[StringIO], Iterable[Any]],
    exception: type[Exception],
    exception_handler: Callable[[Position, Exception], None] | None
) -> TokenMatcher:
    def match(prefix: Slice, rest: Slice | None) -> tuple[Slice, Slice | None]:
        stream = StringIO()
        pos = prefix.end
        rest_copy = rest
        while rest is not None:
            line, rest = rest.to_eol()
            if line is not None:
                pos = line.end
                stream.write(line.get())
                stream.seek(0)
                try:
                    for _ in fn(stream):
                        pass
                except exception as e:
                    if exception_handler:
                        exception_handler(pos, e)
                else:
                    break
            if rest is not None:
                eol, rest = rest.consume_eol()
                assert eol is not None
                pos = eol.end
                stream.write(eol.get())
        else:
            raise LexerError(pos[:2], 'Unexpected EOF')

        assert rest_copy is not None
        return rest_copy[:stream.tell()], rest

    return match

def match_to_eol(prefix: Slice, rest: Slice | None) -> tuple[Slice, Slice | None]:
    out = prefix
    while rest is not None:
        line, rest = rest.to_eol()
        if line is None:
            return out, rest
        out += line

        value = line.get()
        i = len(value) - 1
        for i in range(len(value) - 1, -1, -1):
            if value[i] != '\\':
                break
        num_slashes = len(value) - 1 - i

        if num_slashes % 2 == 0:
            return out, rest

        if rest:
            eol, rest = rest.consume_eol()
            assert eol is not None
            out += eol
    else:
        raise LexerError(out.end[:2], 'Unexpected EOF')

def match_simple(prefix: Slice, rest: Slice | None) -> tuple[Slice, Slice | None]:
    return prefix, rest

def check_tokenize_error(pos: Position, e: Exception) -> None:
    if not isinstance(e, tokenize.TokenError) or not e.args[0].startswith('EOF in multi-line'):
        raise LexerError(pos[:2], f'Error reading Python: {e}')

_SPECS = [
    ('LINECOMMENT', r'#|//', match_to_eol),
    ('PYTHON', r'!\s*',
        match_to_exception(lambda s: tokenize.generate_tokens(s.readline), tokenize.TokenError,
                           check_tokenize_error)),
    ('SHELL', r'>\s*', match_to_exception(lambda s: shlex.split(s.read()), ValueError, None)),
    ('HTTP', r'(?:GET|HEAD|POST|PUT|DELETE|PATCH|OPTIONS) ', match_to_eol),
    ('HEADER', r'-\s*[A-Za-z_-]+\s*:', match_to_eol),
    ('ASSIGN', r'[A-Za-z_][A-Za-z0-9_]*\s*=', match_to_eol),
    ('COMMENT', r'/\*(?:[^*]|\*[^/])+\*/', match_to_eol),
    ('SPACE', r'\s+', match_simple),
    ('COLON', r':', match_simple),
    ('ARROW', r'->', match_simple),
    ('QMARK', r'\?', match_simple),
    ('LBRACKET', r'{', match_simple),
    ('RBRACKET', r'}', match_simple),
    ('LSQUARE', r'\[', match_simple),
    ('RSQUARE', r'\]', match_simple),
    ('DOT', r'\.', match_simple),
    ('DASH', r'-', match_simple),
    ('COMMA', r',', match_simple),
    ('INTEGER', r'-?(?:0|[1-9][0-9]*)', match_simple),
    ('FLOAT', r'-?(?:0|[1-9][0-9]*)(?:\.[0-9]*)?(?:[eE][+-]?[0-9]+)?', match_simple),
    ('TRUE', r'true', match_simple),
    ('FALSE', r'false', match_simple),
    ('NULL', r'null', match_simple),
    ('STRING', r'"(?:[^"\\]|\\(?:["\\/bfnrt]|u[0-9A-Fa-f]{4}))*"', match_simple),
    ('IDENTIFIER', r'[A-Za-z0-9_-]+', match_simple),
]
SPECS = [(name, re.compile(r), matcher) for name, r, matcher in _SPECS]

def simple_lex(data: str) -> Iterator[Token]:
    if not data:
        raise LexerError((0, 0), 'Empty input')

    rest = Slice.from_str(data)
    while rest is not None:
        for name, prefix_regex, matcher in SPECS:
            m = prefix_regex.match(rest.raw, rest.start.offset)
            if m is not None:
                start = m.start() - rest.start.offset
                end = m.end() - rest.start.offset

                prefix = rest[start:end]
                if prefix.end == rest.end:
                    rest = None
                else:
                    rest = rest[end:]

                out, rest = matcher(prefix, rest)
                yield Token(name, out.get(), out.start[:2], out.end[:2])
                break
        else:
            raise LexerError(rest.start[:2], 'No match found')

def lex(data: str) -> Iterator[Token]:
    for token in simple_lex(data):
        if token.type not in ('LINECOMMENT', 'COMMENT', 'SPACE'):
            yield token

if __name__ == '__main__':
    import sys
    with open(sys.argv[1], 'rt') as f:
        for token in lex(f.read()):
            print(token)



import re
from typing import Any, Iterator, TypeVar

from funcparserlib.lexer import Token
from funcparserlib.parser import many, tok, finished, forward_decl, maybe, Parser

from .statement import (Statement, PythonStatement, ShellStatement, HTTPStatement, Identifier,
                        RequestTemplate, RequestTemplateArray, RequestTemplateObject, Placeholder,
                        OutputIdentifier, ResponseTemplate, ResponseTemplateObject, ResponseTemplateArray,
                        ResponseTemplateArrayHandler, OptionalOutputIdentifier)

WHITESPACE = re.compile(r'\s+')

T = TypeVar('T')

def make_none(_: Any) -> None:
    return None

def flatten_many(args: tuple[list[T], list[list[T]]]) -> list[T]:
    return args[0] + sum(args[1], [])

def as_list(value: T) -> list[T]:
    return [value]

def process_assign(assign: str) -> PythonStatement:
    name, value = assign.split('=', 1)
    return PythonStatement.from_assignment(name.strip(), value.strip())

def process_headers(headers: list[str]) -> dict[str, str]:
    rv: dict[str, str] = {}
    for header in headers:
        # - Key: Value
        header = header.split('-', 1)[1].strip()
        key, value = (x.strip() for x in header.split(':', 1))
        rv[key] = value
    return rv

def process_bool(value: str) -> bool:
    return value == 'true'

JSONPath = list[str | int]

def process_json_path_obj(args: tuple[JSONPath, JSONPath]) -> JSONPath:
    return args[0] + args[1]

def process_json_path_array(args: tuple[JSONPath, int, JSONPath]) -> JSONPath:
    return args[0] + [args[1]] + args[2]

def process_jsonlike_object(fields: list[tuple[JSONPath, Any]]) -> Any:
    rv: Any = {}
    for key, value in fields:
        pos = rv
        for cur, nxt in zip(key[:-1], key[1:]):
            if isinstance(cur, int) and cur >= len(pos):
                pos += [Placeholder] * (cur - len(pos) + 1)
            if isinstance(cur, int) or cur not in pos:
                pos[cur] = {} if isinstance(nxt, str) else []
            pos = pos[cur]
        last = key[-1]
        if isinstance(last, int) and last >= len(pos):
            pos += [Placeholder] * (last - len(pos) + 1)
        pos[last] = value
    return rv

def make_http_statement(
        args: tuple[str, dict[str, str], RequestTemplateObject | RequestTemplateArray | str | None,
                    ResponseTemplateObject | ResponseTemplateArray | OutputIdentifier]) -> HTTPStatement:
    method, url = (p.strip() for p in WHITESPACE.split(args[0], 1))
    return HTTPStatement(method, url, *args[1:])

def parse(tokens: Iterator[Token]) -> Iterator[Statement]:
    statement: Parser[Token, Statement] = forward_decl()

    assign_statement = tok('ASSIGN') >> process_assign
    python_statement = tok('PYTHON') >> PythonStatement
    shell_statement = tok('SHELL') >> ShellStatement

    headers = many(tok('HEADER')) >> process_headers

    path_id = tok('IDENTIFIER') >> as_list
    array_unit_path = -tok('LSQUARE') + tok('INTEGER') + -tok('RSQUARE') >> int >> as_list
    array_path = path_id + many(array_unit_path) >> flatten_many
    json_path = array_path + many(-tok('DOT') + array_path) >> flatten_many

    value_number = tok('INTEGER') >> int | tok('FLOAT') >> float
    value_str = tok('STRING') >> str
    value_bool = (tok('TRUE') | tok('FALSE')) >> process_bool
    value_null = tok('NULL') >> make_none
    value_id = tok('IDENTIFIER') >> Identifier
    value = value_number | value_str | value_bool | value_null | value_id

    req_value: Parser[Token, RequestTemplate] = forward_decl()
    req_value_object_field = json_path + -tok('COLON') + req_value >> as_list
    req_value_object: Parser[Token, RequestTemplateObject] = \
            -tok('LBRACKET') + \
            req_value_object_field + many(-tok('COMMA') + req_value_object_field) + \
            -tok('RBRACKET') >> flatten_many >> process_jsonlike_object
    req_value_array: Parser[Token, RequestTemplateArray] = \
            -tok('LSQUARE') + \
            (req_value >> as_list) + many(-tok('COMMA') + req_value) + \
            -tok('RSQUARE') >> flatten_many # type: ignore
    req_value.define(req_value_object | req_value_array | value) # type: ignore

    http_body = req_value_object | req_value_array | (tok('STRING') >> eval)

    response: Parser[Token, ResponseTemplate] = forward_decl()
    response_object_check = json_path + -tok('COLON') + response
    response_object_assign1 = json_path + -tok('QMARK') + -tok('ARROW') + \
                              (tok('IDENTIFIER') >> OptionalOutputIdentifier)
    response_object_assign2 = json_path + -tok('ARROW') + (tok('IDENTIFIER') >> OutputIdentifier)
    response_object_field = (response_object_check | response_object_assign1 |
                             response_object_assign2) >> as_list
    response_object: Parser[Token, ResponseTemplateObject] = \
            -tok('LBRACKET') + \
            response_object_field + many(-tok('COMMA') + response_object_field) + \
            -tok('RBRACKET') >> flatten_many >> process_jsonlike_object
    response_handler = \
            -tok('LSQUARE') + \
            response + many(statement) + \
            -tok('RSQUARE') >> ResponseTemplateArrayHandler.make
    response_array: Parser[Token, ResponseTemplateArray] = \
            -tok('LSQUARE') + \
            response + many(-tok('COMMA') + response) + \
            -tok('RSQUARE') >> flatten_many # type: ignore
    response_output_id = -tok('ARROW') + tok('IDENTIFIER') >> OutputIdentifier
    response.define(response_object | response_handler | response_array | response_output_id | value) # type: ignore

    http_block: Parser[Token, HTTPStatement] = \
            tok('HTTP') + headers + maybe(-tok('DASH') + http_body) + response \
            >> make_http_statement # type: ignore

    statement.define(assign_statement | python_statement | shell_statement | http_block) # type: ignore

    document = many(statement) + finished

    return iter(document.parse(list(tokens))[0])

if __name__ == '__main__':
    import sys
    from .lexer import lex
    with open(sys.argv[1], 'rt') as f:
        for x in parse(lex(f.read())):
            print(x)



from setuptools import setup
from pathlib import Path

setup(
    name='api-query',
    version='0.2.0',
    description='Tool to generate and run a script to navigate a REST API.',
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url='https://github.com/es1024/api-query',
    license='MIT',
    packages=['api_query'],
    package_data={'api_query': ['py.typed', '__init__.pyi']},
    install_requires=[
        'aiohttp',
        'funcparserlib>=1.0.0',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Communications',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    entry_points={
        'console_scripts': [
            'api-query = api_query.main:main',
        ],
    },
)



"""
    Helper functions used by the generated script.
"""

import asyncio
import time
import importlib
import logging
import sys
from types import TracebackType
from typing import NamedTuple


LOG = logging.getLogger('query')
LOG.setLevel(logging.INFO)
if not LOG.handlers:
    formatter = logging.Formatter(f"%(levelname)8s: %(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    LOG.addHandler(stream_handler)

logging.addLevelName(logging.FATAL, 'fatal')
logging.addLevelName(logging.ERROR, 'error')
logging.addLevelName(logging.WARNING, 'warning')
logging.addLevelName(logging.INFO, 'info')
logging.addLevelName(logging.DEBUG, 'debug')

def set_log_level(level: str) -> None:
    LOG.setLevel(level)
    stream_handler.setLevel(level)


class ConcurrencyLimit:
    def __init__(self, max_concurrent: int) -> None:
        self.semaphore = asyncio.BoundedSemaphore(max_concurrent)

    async def __aenter__(self) -> None:
        await self.semaphore.acquire()

    async def __aexit__(
            self, _exc_type: type[BaseException] | None, _exc_val: BaseException | None,
            _exc_tb: TracebackType | None) -> None:
        self.semaphore.release()


class RateLimit:
    def __init__(self, max_qps: int) -> None:
        self.token_interval = 1.0 / max_qps
        self.last_token_time = 0
        self.max_tokens = max_qps
        self.tokens = max_qps
        self.lock = asyncio.Lock()
        self.semaphore = asyncio.BoundedSemaphore(self.tokens)

    async def _update_tokens(self) -> None:
        while True:
            async with self.lock:
                now = time.time()
                if self.tokens < self.max_tokens and now - self.last_token_time > self.token_interval:
                    new_tokens = min(self.max_tokens - self.tokens,
                                     int((now - self.last_token_time) / self.token_interval))
                    self.tokens += new_tokens
                    if self.tokens == self.max_tokens:
                        self.last_token_time = now
                    else:
                        self.last_token_time += new_tokens * self.token_interval
                    for _ in range(new_tokens):
                        self.semaphore.release()
            await asyncio.sleep(self.token_interval)

    async def get_token(self) -> None:
        update_task = asyncio.create_task(self._update_tokens())
        acquire_task = asyncio.create_task(self.semaphore.acquire())
        finished, unfinished = await asyncio.wait({update_task, acquire_task},
                                                  return_when=asyncio.FIRST_COMPLETED)
        for task in finished:
            task.result()
        assert acquire_task in finished
        for task in unfinished:
            task.cancel()
        async with self.lock:
            self.tokens -= 1


def resolve_imports() -> None:
    frame = sys._getframe().f_back # type: ignore[reportPrivateUsage]
    assert frame is not None
    for name in frame.f_code.co_names:
        if name not in frame.f_globals and \
                name not in frame.f_locals and \
                name not in dir(__builtins__):
            try:
                frame.f_globals[name] = importlib.import_module(name)
                LOG.debug(f'Imported {name}')
            except ModuleNotFoundError:
                pass


async def run_shell(command: str) -> None:
    LOG.debug(f'Running `{command}`')
    proc = await asyncio.create_subprocess_shell(command)
    await proc.communicate()
    if proc.returncode != 0:
        raise RuntimeError(f'`{command}` failed with return code {proc.returncode}')


class RetryConfig(NamedTuple):
    retry_count: int
    base_delay: float
    max_delay: float
